import { Metadata, Http } from '@zeedhi/core';
import { Modal, ModalService, Form, DialogService, LoadingService, Select, SelectMultiple, IFormComponent} from '@zeedhi/common';
import { TekGrid } from "@zeedhi/teknisa-components-common";
import { CustomAlertService } from '@/Services/CustomAlertService';
import { Utilities } from '@/Services/Utilities';
import { ExpressionInterpreter } from '@/Services/ExpressionInterpreter';
import { Highlight } from '@zeedhi/zd-highlight-common';
import Prism from 'prismjs';
import 'prismjs/components/prism-markup-templating.min';
import 'prismjs/components/prism-php.min';

export class DataSourceController {
    public isGridVisible: any = true;
    public get isDetailVisible() {
        return !this.isGridVisible;
    }
    public showGrid() {
        this.isGridVisible = true;
    }
    public showDetail() {
        this.isGridVisible = false;
    }
    //  Referente ao componente Highlight
    public get addDatasourceQuery() {
        const queries = Metadata.getInstances('query');
        return queries.find((query: any) => query.parent.name === 'modalDataSourceForm');
    }
    public get editDatasourceQuery() {
        const queries = Metadata.getInstances('query');
        return queries.find((query: any) => query.parent.name === 'dataSourceForm');
    }
    public loadLanguage = this.loadLanguageAux.bind(this);
    public loadLanguageAux() {
        const query = this.isGridVisible ? this.addDatasourceQuery : this.editDatasourceQuery;
        if (query) {
            const { code, language } = query;
            return Prism.highlight(code, Prism.languages[language], language);
        }
        return "";
    }
    public loadLanguageExpression(params: any) {
        const { code, language } = Metadata.getInstance<Highlight>('expression');
        return Prism.highlight(code, Prism.languages[language], language);
    }
    // Métodos de Dialog e Modal
    public showDeleteDialog(params: any) {
        Utilities.genericBeforeDeleteTekGrid(params, "/dataSource/remove", "datasource-grid");
    }

    // Coluna
    private columnModal!: Modal;
    public columnModalData: any;
    public columnModalType: string = "";
    public get isColumnFact() {
        const form = Metadata.getInstance<Form>("columnForm");
        return form.value.type === "fact";
    }
    public get isColumnDimension() {
        const form = Metadata.getInstance<Form>("columnForm");
        return form.value.type === "dimension";
    }
    public get isColumnFactAndNotGenerated() {
        return this.isColumnFact && !this.columnModalData?.isGenerated;
    }
    public get isColumnUsingCase() {
        const form = Metadata.getInstance<Form>("columnForm");
        return this.isColumnFact && form.value.specialFunction === "case";
    }
    public get isColumnNotUsingCase() {
        const form = Metadata.getInstance<Form>("columnForm");
        return !this.isColumnFact || form.value.specialFunction !== "case";
    }
    public get isColumnUsingModule() {
        const form = Metadata.getInstance<Form>("columnForm");
        return this.isColumnFact && form.value.specialFunction === "mod";
    }
    public get isColumnUsingPower() {
        const form = Metadata.getInstance<Form>("columnForm");
        return this.isColumnFact && form.value.specialFunction === "power";
    }
    public async openColumnModal({ row }: any) {
        this.columnModal?.destroy();
        this.columnModal = await ModalService.createFromJson('dataSource/columnModal/columnModal', true);
        this.columnModalData = row === undefined ? {type: 'fact', datatype: 'A', direction: 'asc', relatedFactMultiple: []} : row;
        this.columnModal.show();
    }
    public closeColumnModal() {
        this.columnModal.hide();
        this.columnModal.destroy();
    }
    public columnFormMounted({ component }: any) {
        const expression = Metadata.getInstance<Highlight>('expression');
        const relatedFacts = ['relatedFactMultiple', 'relatedFact', 'secondRelatedFact'];
        relatedFacts.forEach((relatedFactName: any) => {
            const relatedFact = Metadata.getInstance<SelectMultiple>(relatedFactName);
            relatedFact.datasource.updateData(this.datasourceData.columns.filter((col: any) => col.isFact));
            relatedFact.value = this.columnModalData[relatedFactName] || [];
        });
        component.value = this.columnModalData;
        expression.code = component.value.expression || "";
    }
    public columnSpecialFunctions: any = {};
    public columnFormGrid = { datatype: 3, color: 6};
    public updateColumnConditionalFormats() {
        const direction = Metadata.getInstance<Select>('direction');
        const specialFunction = Metadata.getInstance<Select>('specialFunction');
        const relatedFact = Metadata.getInstance<Select>('relatedFact');
        const relatedFact2 = Metadata.getInstance<Select>('secondRelatedFact');
        const potentialExpoent = Metadata.getInstance<Select>('potentialExpoent');

        if (this.isColumnFact) {
            direction.addValidation('required', {});
            this.columnFormGrid.datatype = 6;
            this.columnFormGrid.color = 3;
        } else {
            direction.removeValidation('required');
            this.columnFormGrid.datatype = 3;
            this.columnFormGrid.color = 6;
        }

        const requiresRelatedFact = ['abs', 'sign', 'ceil', 'floor', 'sqrt', 'mod', 'power'];
        if (specialFunction.value === 'power') {
            potentialExpoent.addValidation('required', {"message": "Por favor, preencha o expoente"});
        } else {
            potentialExpoent.removeValidation('required');
        }

        if (specialFunction.value === 'module') {
            relatedFact2.addValidation('required', {"message": "Por favor, preencha o segundo fato relacionado para a função mod"});
        } else {
            relatedFact2.removeValidation('required');
        }

        if (requiresRelatedFact.includes(specialFunction.value)) {
            relatedFact.addValidation('required', {"message": 'Por favor, preencha o fato relacionado para a função ' + specialFunction.value});
        } else {
            relatedFact.removeValidation('required');
        }
    }
    public async importColumns() {
        const gridColumns = Metadata.getInstance<TekGrid>("gridColumns");
        try {
            gridColumns.datasource.loading = true;
            const res = await Http.post("/dataSource/syncColumns", {
                requestType: "Row",
                row: this.datasourceData || {}
            });
            gridColumns.datasource.loading = false;
            if (res.data.messages?.length > 0) {
                const message = res.data.messages[0].message;
                DialogService.show({
                  name: 'warning-dialog',
                  type: 'warning',
                  title: ' ',
                  text: message
                });
                if (res.data.messages[0].type === 'E') {
                    return;
                }
            }
            const columns = res.data.dataset && res.data.dataset.columns ? res.data.dataset.columns : [];
            this.datasourceData.columns = columns;
            this.columnGridChange({component: gridColumns});
            gridColumns.datasource.loading = false;
        } catch (err) {
            gridColumns.datasource.loading = false;
            CustomAlertService.errorAlert("Ocorreu um erro.");
            console.error(err);
        }
    }

    public selectColumn({ row, component }: any) {
        if (row?.isGenerated) {
            DialogService.show({
              name: 'warning-column-select',
              type: 'warning',
              title: ' ',
              text: 'Não é possivel selecionar uma coluna gerada!'
            });
            component.selectedRows = component.selectedRows.filter((value: any) => !value.isGenerated);
        }
    }
    public async deleteColumn({ component }: any) {
        const columnsToDelete = component.selectedRows.filter((value: any) => !value.isGenerated);
        if (columnsToDelete.length === 0) {
            return;
        }
        this.datasourceData.columns = this.datasourceData.columns.filter((column: any) => !columnsToDelete.includes(column));
        const res = await this.saveColumns(this.datasourceData);
        const columnGrid = Metadata.getInstance<TekGrid>("gridColumns");
        columnGrid.datasource.updateData(res.columns);
        CustomAlertService.deleteSuccessAlert();
    }

    public testExpression() {
        let expression: any = "";
        const formValues = Metadata.getInstance<Form>("columnForm").value;
        const expressionEl = Metadata.getInstance<Highlight>("expression");
        try {
            if (formValues.specialFunction === "case") {
                expression = this.readCaseExpression(expressionEl.code, this.datasourceData.hasDataMining, formValues);
            } else {
                expression = this.readRegularExpression(formValues.name, expressionEl.code);
            }
            if (expression || expressionEl.code) {
                if (expression) {
                    CustomAlertService.successAlert("Expressão válida.");
                } else {
                    CustomAlertService.errorAlert("Expressão inválida.");
                }
            }
            return expression;
        } catch (e) {
            console.error(e);
            CustomAlertService.errorAlert(String(e));
        }
        return undefined;
    }
    public readRegularExpression(name: string, code: string) {
        const data = this.datasourceData.columns;
        const facts = data.filter((column: any) => column.isFact === true).map((column: any) => column.name);
		const nonFacts = data.filter((column: any) => column.isFact === false).map((column: any) => column.name);
		return ExpressionInterpreter.readExpression(code, facts, nonFacts, name);
    }
    public readCaseExpression(expression: string, hasDataMining: any, formValues: any) {
		if (expression) {
            if (!hasDataMining) {
                if (this.checkExpressionString(expression.toUpperCase())) {
					return expression;
                }
                CustomAlertService.errorAlert("Para criar uma expressão case sem mineração de dados, você deve consultar a sintaxe \"CASE\" de seu banco de dados.");
            } else {
                const caseExp = this.formatCaseExp(expression);
                const regEx = new RegExp("(CASE.* [^ ].* [<>!=].* [^ ].* (THEN).* [^ ].*).* DEFAULT.* ([^ ].*)");
                const testExp = regEx.test(caseExp);
                if (testExp) {
                    formValues.caseExp = caseExp;
                    return expression;
                }
                CustomAlertService.errorAlert("Para expressões CASE, siga o modelo: CASE [condição] THEN [expressão] DEFAULT [expressão]\n\nEx: CASE $fatoRelacionado1 != 0 THEN (( $fatoRelacionado2 - $fatoRelacionado1 )*100 / $fatoRelacionado1 ) DEFAULT 100");
            }
		} else {
            CustomAlertService.errorAlert("O campo \"expressões\" é obrigatório para funções CASE.");
		}
		return "";
    }
    public checkExpressionString(expression: string) {
		const arrayKeyWords = expression.split(" ");
		const arrayExpression = expression.split("");
		let checkExpression;
		if (arrayKeyWords[0] === "CASE") {
            let countChar = 0;
            arrayExpression.forEach((val: string) => {
                if (val === '(') {
					countChar += 1;
				} else if (val === ')') {
					countChar -= 1;
					if (countChar < 0) {
						CustomAlertService.errorAlert("Não há \"(\" para fechar com \")\".");
						checkExpression = false;
					}
				}
            });
			if (countChar > 0) {
				CustomAlertService.errorAlert("Existem mais \"(\" do que o necessário.");
				checkExpression = false;
			} else if (countChar < 0) {
				CustomAlertService.errorAlert("Existem mais \")\" do que o necessário.");
				checkExpression = false;
			} else if (countChar === 0 && checkExpression !== false) {
				checkExpression = true;
				CustomAlertService.errorAlert("Por favor, verifique a sintaxe da expressão case do banco de dados usado.");
			}
		} else {
			checkExpression = false;
		}
		return checkExpression;
	}
	public formatCaseExp(expression: string) {
        let newExp = '';
		const words = expression.split(' ');

		words.forEach((word: string) => {
			if (word[0] !== '$') {
				word = word.toUpperCase();
			}

			word = word.replace('\n', ' ');
			newExp += word + ' ';
		});

		return newExp;
	}

    public async saveColumnModal() {
        const form = Metadata.getInstance<Form>("columnForm");
        const query = Metadata.getInstance<Highlight>('expression');
        const values = {...form.value};
        const queryCode = query.code ? this.parseQueryCode(query.code) : "";
        if (!form.validate() || (this.isColumnFactAndNotGenerated && this.isColumnUsingCase && !queryCode.trim())) {
            if (!queryCode.trim()) {
                CustomAlertService.errorAlert("O campo de código é obrigatório e deve ser preenchido");
            }
            return;
        }
        const aux = {
            ...this.columnModalData,
            ...values,
            expression: queryCode,
            isDate: values.datatype === 'D',
            isDimension: values.type === 'dimension',
            isFact: values.type === 'fact',
            isGenerated: this.columnModalData.isGenerated || false
        };
        if (aux.color && !Array.isArray(aux.color)) {
            aux.color = [aux.color];
        }
        let ignoreAttrs: string[] = [];
        if (aux.isFact) {
            ignoreAttrs = aux.isGenerated ? ["isKey", "isFunctionAccumulative", "specialFunction", "relatedFactLabel", "expression", "validateExp"] : ["isKey"];
        } else if (aux.isDimension) {
            ignoreAttrs = ["function", "isFunctionAccumulative", "specialFunction", "relatedFactLabel", "expression", "validateExp", "mask"];
        }
        ignoreAttrs.forEach((key: string) => { delete aux[key]; });
        // Caso a coluna já existisse, substitui a posição pelas novas edições na coluna, caso o contrario, a insere no vetor
        const oldPos = this.datasourceData.columns.findIndex((col: any) => col === this.columnModalData);
        if (oldPos >= 0) {
            this.datasourceData.columns[oldPos] = aux;
        } else {
            this.datasourceData.columns.push(aux);
        }
        this.columnModalData = aux;
        let res = null;
        if (this.columnModalData.isDimension) {
            res = await this.saveColumn(this.datasourceData, this.columnModalData);
        } else if (this.datasourceData.hasDataMining) {
            if (queryCode !== "") {
                if (this.testExpression() !== undefined || this.readRegularExpression(this.columnModalData.name, queryCode) !== undefined) {
					res = await this.saveColumn(this.datasourceData, this.columnModalData);
				} else {
					CustomAlertService.errorAlert("A coluna não pode ser salva.");
				}
            } else {
                res = await this.saveColumn(this.datasourceData, this.columnModalData);
            }
        } else if (this.readCaseExpression(this.columnModalData.expression, false, this.columnModalData) !== undefined || this.readRegularExpression(this.columnModalData.name, queryCode) !== undefined) {
			res = await this.saveColumn(this.datasourceData, this.columnModalData);
		} else {
			CustomAlertService.errorAlert("A coluna não pode ser salva.");
		}

        if (res) {
            const columnGrid = Metadata.getInstance<TekGrid>("gridColumns");
            columnGrid.datasource.updateData(res.columns);
            this.closeColumnModal();
        }
    }
    public saveColumn(datasourceData: any, columnData: any) {
		columnData._altered = true;
		if (!columnData.isGenerated && columnData.expression) {
			let expression = columnData.expression;
			const beginningExpression = expression.substring(0, 4).toUpperCase();
			if (beginningExpression !== "CASE") {
                expression = this.readRegularExpression(columnData.name, expression);
			}
			columnData.functions = expression.functions;
			columnData.mathExpressions = expression.mathExpressions;
		}
		return this.saveColumns(datasourceData);
    }
    public async saveColumns(datasourceData: any) {
        datasourceData._columnsChanged = true;
        datasourceData.columns = datasourceData.columns.map((column: any) => ({
            ...column,
            isSortable: !!column.direction,
            isFact: column.type === 'fact',
            isDimension: column.type === 'dimension'
		}));
		datasourceData.keys = datasourceData.columns.filter((column: any) => column.isKey).map((currentValue: any) => currentValue.name);
		const res = await this.saveDataSource(datasourceData);
		return {res, columns: datasourceData.columns};
    }

    // TRECHO DE CODIGO RELACIONADO AO MASTERDETAIL, GRID E FORM DO DATASOURCE
    private modal!: Modal;
    public closeDetails() {
        if (this.isDetailVisible) {
            this.showGrid();
        } else {
            this.modal.hide();
        }
    }
    public async createAddModal() {
        this.modal = await ModalService.createFromJson('dataSource/AddModal/AddModal', true);
    }

    protected datasourceData: any;
    public detailTitleText: string = '';
    public switchToDetail({ component, row }: any) {
        if (component.editing === true) return;
        this.datasourceData = row;
        this.detailTitleText = this.datasourceData?.name || "Fonte de Dados";
        this.showDetail();
        this.fillDataSourceForm({ row });
        const columnGrid = Metadata.getInstances("gridColumns");
        if (columnGrid.length > 0 && columnGrid[0]) {
            this.columnGridChange({ component: columnGrid[0] });
        }
    }
    public async openModal({ component, row }: any) {
        this.modal?.destroy();
        await this.createAddModal();
        this.datasourceData = {joinDataSource: "N", hasDataMining: "S", willUpdateIndicators: "N", isScheduleActive: "S"};
        this.modal.show();
    }
    public columnGridChange({ component }: any) {
        const columns = this.datasourceData.columns || [];
        component.datasource.updateData(columns);
    }
    // Métodos da query
    public extractStringsRangeFromQuery(code: string) {
        const stringRegex = /(["'])(?:(?=(\\?))\2.)*?\1/g;
        return [...code.matchAll(stringRegex)].map((codeString: any) => [codeString.index, codeString.index + codeString[0].length]);
    }

    public strReplaceNotIn(forbideenRange: any, inForbideenFunc: any, notInForbideenFunc: any) {
        return (...params: any) => {
            const commentStart = params[params.length - 2];
            const containingString = forbideenRange.find(([start, end]: any) => commentStart > start && commentStart < end);
            if (containingString) {
                return inForbideenFunc(params, containingString);
            }
            return notInForbideenFunc(params);
        };
    }

    public clearComments(code: string): string {
        if (code === "") {
            return "";
        }
        const commentRegex = /(--.*)|(((\/\*)+?[\w\W]+?(\*\/)+))/g;
        const stringRanges = this.extractStringsRangeFromQuery(code);
        return code.replace(commentRegex, this.strReplaceNotIn(stringRanges, (params: any, [start, end]: any) => {
            const comment = params[0];
            const onString = comment.slice(0, end - start);
            const notOnString = comment.slice(end - start - 1, comment.length);
            return onString + this.clearComments(notOnString);
        }, () => ""));
    }

    public parseQueryCode(code: string) {
        const doubleWhiteSpaceRegex = /\s+/g;
        const commentless = this.clearComments(code);
        const stringRanges = this.extractStringsRangeFromQuery(commentless);
        const parsedCode = commentless.replace(doubleWhiteSpaceRegex, this.strReplaceNotIn(stringRanges, (params: any) => params[0], () => " "));
        return parsedCode.trim();
    }
    //  Metodos do Form
    public fillDataSourceForm({ row }: any) {
        if (!row) {
            return;
        }
        const values = row || this.datasourceData;
        const form = Metadata.getInstance<Form>("dataSourceForm");
        const query = this.isGridVisible ? this.addDatasourceQuery : this.editDatasourceQuery;
        const accessProfiles = form.getChildInstance<SelectMultiple>("accessProfiles");
        form.value = values;
        query.code = values.query_view || values.query || "";
        accessProfiles.setValue(values.accessProfiles || []);
    }

    public async saveDatasourceForm() {
        const form = this.isGridVisible ? Metadata.getInstance<Form>("modalDataSourceForm") : Metadata.getInstance<Form>("dataSourceForm");
        const query = this.isGridVisible ? this.addDatasourceQuery : this.editDatasourceQuery;
        const values = form.value;
        const connection = values.connection;
        const queryCode = query.code ? this.parseQueryCode(query.code) : "";
        if (!form.validate() || !queryCode.trim()) {
            if (!queryCode.trim()) {
                CustomAlertService.errorAlert("O campo de código é obrigatório e deve ser preenchido");
            }
            return;
        }
        const data = {
            ...values,
            changed: true,
            columns: this.datasourceData.columns || [],
            connection: connection?._id,
            connectionName: connection?.name,
            room: connection?.room,
            createdBy: this.datasourceData.userId,
            importStatus: this.datasourceData.importStatus || "",
            originDataSources: this.datasourceData.originDataSources || "",
            query: queryCode,
            query_view: query.code || "",
            relations: this.datasourceData.relations || [],
            _queryChanged: this.datasourceData._id ? !(this.datasourceData.query === queryCode) : true,
            _id: this.datasourceData?._id
        };
        try {
            const res = await this.saveDataSource(data);
            CustomAlertService.saveSuccessAlert();
            if (res.data.messages?.length > 0) {
                const message = res.data.messages[0].message;
                DialogService.show({
                  name: 'warning-dialog',
                  type: 'error',
                  title: ' ',
                  text: message
                });
            }
            this.reloadGrid();
            this.closeDetails();
        } catch (err) {
            LoadingService.hide();
            CustomAlertService.errorAlert("Ocorreu um erro.");
            console.error(err);
        }
    }
    public async saveDataSource(datasourceData: any) {
        LoadingService.show();
        const res = await Http.post("/dataSource/save", {
            requestType: "DataSet",
            disableLoader: false,
            dataset: [datasourceData._id ? datasourceData : {
                ...datasourceData,
                __createdLocal: true,
                __is_new: true,
                createdBy: this.user.userId
            }]
        });
        LoadingService.hide();
        return res;
    }

    protected user: any;
    public async updateLoggedUserInfo() {
        Http.post("/getLoggedUser", {nrorg: null, nrorgPrincipal: null}).then((res) => {
            const login = res.data;
            this.user = {
                userId: login.userId,
                username: login.username
            };
        });
    }

    public handleConnectionResponse({ component }: any) {
        const info = component?.currentRow;
        if (info?.WALLET_NRORG === undefined) {
            return;
        }
        const {data, WALLET_NRORG, pagination} = info;
        component.data = data;
        component.total = pagination.total;
        component.limit = pagination.limit;
        component.page = pagination.page;
        component.currentRow = data;
    }
    public reloadGrid() {
        const grid = Metadata.getInstance<TekGrid>("datasource-grid");
        grid.reload();
    }
    public async saveInlineEdit(params: any) {
        Utilities.genericBeforeSaveTekGrid(params, "dataSource/save", "datasource-grid");
    }

    public adjustFilter({ component }: any) {
        const datasource = component.datasource;
        const filter = datasource.dynamicFilter;
        Object.entries(filter).forEach(([k, filters]: any) => {
            filter[k] = filters.filter(({value}: any) => !(value instanceof Array && value.length === 0));
            if (filter[k].length === 0) {
                delete filter[k];
            }
        });
    }
}
